# -*- coding: utf-8 -*-
"""
This module contains constants that are needed in multiple modules.

:copyright: (c) 2012 by Rhett Garber.
:license: ISC, see LICENSE for more details.
"""

MAX_BATCH_WRITE_ITEMS = 25
MAX_BATCH_READ_ITEMS = 100

